import os
# import rasterio

from . import log


def to_tiff(filename_in, filename_out, format_in, cog=False):
    """
    For accepted formats (in): https://gdal.org/drivers/raster/index.html
    """
    import rasterio

    format_out = 'GTiff'
    format_in = 'ISIS3' if format_in == 'ISIS' else format_in

    try:
        if cog:
            res = tiff2cog(filename_in, filename_out)
        else:
            src = rasterio.open(filename_in, 'r', driver=format_in)
            data = src.read()
            params = src.meta
            params['driver'] = format_out
            dst = rasterio.open(filename_out, 'w', **params)
            dst.write(data)
            dst.close()
            src.close()
    except Exception as err:
        raise err
        return None
    return filename_out


def merge(filenames, output):
    """
    Return filename of merged 'filenames' GeoTIFFs

    Input:
        filenames : list
            List of filenames to merge
        output : string
            Mosaic filename
    """
    import rasterio
    from rasterio.merge import merge as riomerge
    log.debug("Running 'merge' method.")

    with rasterio.open(filenames[0]) as src:
        meta = src.meta.copy()

    # The merge function returns a single array and the affine transform info
    arr, out_trans = riomerge(filenames)

    meta.update({
        "driver": "GTiff",
        "height": arr.shape[1],
        "width": arr.shape[2],
        "transform": out_trans
    })

    # Write the mosaic raster to disk
    with rasterio.open(output, "w", **meta) as dest:
        dest.write(arr)

    return output


def rescale(filename_in, filename_out, factor=0.5):
    """
    Resample data for faster processing. Rescale to HALF the resolution by default.
    """
    import rasterio
    from rasterio.enums import Resampling

    with rasterio.open(filename_in) as src:

        height = int(src.height * factor)
        width = int(src.width * factor)
        transform = src.transform * src.transform.scale(
                    (src.width / width),
                    (src.height / height)
        )

        # resample data to target shape
        data = src.read(
            out_shape=(src.count, height, width),
            resampling=Resampling.bilinear
        )
        data[data<=src.nodata] = src.nodata

        # copy src metadata, update as necessary for 'dst'
        kwargs = src.meta.copy()
        kwargs.update({
            'transform': transform,
            'width': width,
            'height': height
        })

        # reproject "src" to "dst"
        with rasterio.open(filename_out, 'w', **kwargs) as dst:
            for i, band in enumerate(data, 1):
                dst.write(band, i)

        return filename_out

resample = rescale



from .bbox import string_2_dict as bbox_string_2_dict

from . import log

# === Inspect functions
# ref: https://fabianlee.org/2019/09/22/python-using-a-custom-decorator-to-inspect-function-arguments/
#
import functools
import inspect

def logintrospect(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        inspect_decorator(func, args, kwargs)
        result = func(*args, **kwargs)
        print("-return " + str(result))
        return result
    return wrapper


def inspect_decorator(func,args,kwargs):
  funcname = func.__name__
  print("function {}()".format(funcname))

  # get description of function params expected
  argspec = inspect.getargspec(func)

  # go through each position based argument
  counter = 0
  if argspec.args and type(argspec.args is list):
    for arg in args:
      # when you run past the formal positional arguments
      try:
        print("*" + str(argspec.args[counter]) + " = " + str(arg))
        counter+=1
      except IndexError as e:
        # then fallback to using the positional varargs name
        if argspec.varargs:
          varargsname = argspec.varargs
          print("*" + varargsname + " = " + str(arg))
        pass

  # finally show the named varargs
  if argspec.keywords:
    kwargsname = argspec.keywords
    for k,v in kwargs.items():
      print("**" + kwargsname + " " + k + " = " + str(v))

# ==========

@logintrospect
def geojson_2_geodataframe(records):
    import geopandas as gpd
    import shapely

    assert isinstance(records, list), "Expected a list [{}], instead got {}".format(type(records))
    gpdrecs = []
    for rec in records:
        geom = rec['geometry']
        if isinstance(geom, str):
            try:
                # rec['geometry'] = shapely.wkt.loads(rec['geometry'])
                geom = shapely.wkt.loads(geom)
            except Exception as err:
                log.error(err)
                raise err
        elif geom is None:
            pass
        else:
            try:
                geom = shapely.geometry.asShape(geom)
            except Exception as err:
                log.error(err)
                raise err
        gpdrec = rec['properties']
        gpdrec['geometry'] = geom
        gpdrecs.append(gpdrec)
    gdf = gpd.GeoDataFrame(gpdrecs)
    return gdf


def products_2_geojson(products, filename):
    """
    Write products to a GeoJSON 'filename'. Return GeoPandas dataframe.

    > Note: This function modifies field 'geometry' from 'products'

    products: list of product records (from search_footprints)
    filename: GeoJSON filename for the output
    """
    # import geopandas as gpd
    # import shapely

    assert isinstance(products, list), "Expected 'products' to be a list"
    assert filename and filename.strip() != '', "Give me a valid filename"

    gdf = geojson_2_geodataframe(products)

    gdf.to_file(filename, driver='GeoJSON')
    print("File '{}' written.".format(filename))
    return

json_2_geojson = products_2_geojson



from npt import log


def bake_container(name):
    """
    Return a Bash login shell from "inside" container 'name'

    "Inside" means it will exec whatever (wrapped) command from inside the
    (docker) container in a Bash login shell.
    """
    from sh import docker
    _exec_ = "exec -t {name!s} bash --login -c"
    if name not in list_containers():
        # TODO: together with an option "autorun", start/run container if not yet.
        raise ValueError
    dsh = docker.bake(_exec_.format(name=name).split())
    return dsh


def containers():
    """
    Return list of container (names) instanciated
    """
    from sh import docker, awk, tail
    from io import StringIO

    buf = StringIO()
    try:
        tail(awk(docker('ps','-a'), '{print $NF}'), '-n+2', _out=buf)
    except Exception as err:
        log.error(err)

    containers = buf.getvalue().split()
    return containers

list_containers = containers


def restart(name):
    """
    Re/Start a container
    """
    from sh import docker
    if name in containers():
        docker('start',name)
        return True
    else:
        log.warning(f"No container '{name}' instantiated. Nothing to restart.")
        return False

container_restart = restart



import os
from pathlib import Path

from npt import log
from ..utils import raster
from ..utils.filenames import change_extension, change_dirname, insert_preext


def mosaic(geojson:dict, basepath:str, prefix_output:str="mosaic",
    method:str='all', scale_factor:float=0.1):
    """
    Make mosaic from files in 'input_geojson' file. Write GeoJSON with mosaic feature.

    Input:
        scale_factor: float
            Between (0:1], define the down-scaling to apply on both axis
    """
    import geopandas
    from datetime import datetime

    _sep = '_'

    # Read the GeoJSON (dict) given as input
    gdf = geopandas.GeoDataFrame.from_features(geojson['features'])
    gdf = gdf.loc[gdf.area.sort_values(ascending=False).index]
    log.info("{:d} features read".format(len(gdf)))

    # Define the list of source images
    filenames = list(gdf['tiff_path'])
    log.debug(f"Sources: {filenames}")

    # Define the final/output geometry (the union of source geometries)
    geometry = [gdf.geometry.unary_union]

    # Define the output properties/metadata
    properties = {}
    for p in gdf.columns:
        # Use fields that have the same value across all sources
        if len(gdf[p].unique()) == 1:
            val = list(gdf[p].unique())
            properties[p] = val
        else:
            val = None

    # Add 'id_sources' for the field of images source IDs as a comma-separated string-list
    properties['id_sources'] = [','.join(list(gdf['id']))]

    # Define a GeoDataFrame/GeoJSON with (one feature of) such metadata
    ngdf = geopandas.GeoDataFrame(properties, geometry=geometry)
    _rec = ngdf.iloc[0]

    # Define name of output file
    c_lon,c_lat = geometry[0].centroid.coords.xy
    c_lon = '{:03d}'.format(int(c_lon[0]))
    c_lat = '{:03d}'.format(int(c_lat[0]))

    dtime = datetime.now().strftime("%Y%m%dT%H%M%S%f")

    mosaic_filename = _sep.join([
                                f"{prefix_output}",
                                f"{_rec['inst'].lower()}",
                                f"{dtime}",
                                f"lon{c_lon}lat{c_lat}.tif"
                                ])

    mosaic_path = _mosaic(filenames, basepath=basepath,
                          mosaic_filename=mosaic_filename,
                          method=method, scale_factor=scale_factor)

    assert mosaic_path

    ngdf['tiff_kbytes'] = [int(os.path.getsize(mosaic_path)/1000)]
    ngdf['tiff_path'] = [mosaic_path]

    return ngdf


def _mosaic(filenames, basepath, mosaic_filename=None, make_dirs=True,
            method='all', scale_factor=0.1):
    """
    * method options: virtual, warp, merge
    """
    # from npt.utils.raster import warp, merge, virtual
    from npt.utils.raster import merge

    # Rescale?
    assert 0 < scale_factor <= 1, "Scale-factor should be between (0:1]"
    if scale_factor < 1:
        filenames_rescaled = []
        for filename in filenames:
            _f = str(scale_factor).replace('.','')
            filename_resc = insert_preext(change_dirname(filename, basepath), f'rescaled{_f}')
            log.debug(f"Rescaled filename to-write: {filename_resc}")
            try:
                filename_resc = raster.rescale(filename, filename_resc, scale_factor)
            except Exception as err:
                log.error(f"Failed to rescale (by {scale_factor}) file: {filename}")
                log.error(str(err))
                continue

            filenames_rescaled.append(filename_resc)

        filenames = filenames_rescaled[:]

    assert len(filenames), "No filenames/source to mosaic(?)!"

    if not os.path.isdir(basepath):
        if make_dirs:
            os.makedirs(basepath, exist_ok=True)
        else:
            print("Path '{}' does not exist.".format(basepath))
            return None

    if not mosaic_filename:
        mosaic_filename = "mosaic.tif"
    output = Path(basepath) / mosaic_filename
    output = output.as_posix()

    output = merge(filenames, output, method=method)

    return output



"""
Some Neanias-Planets Tools for backend planetary data handling
"""
from ._log import log

from . import datasets
from . import search
from . import download
from . import reduce
from . import mosaic
from . import pipelines
from . import utils

from . import _version
__version__ = _version.get_versions()['version']



from .formatters import geojson_2_geodataframe, products_2_geojson, json_2_geojson
from npt.utils import filenames

import json

from . import log

def read(filename):
    """
    Return 'features' from GeoJSON
    """
    with open(filename, 'r') as fp:
        js = json.load(fp)
    #     features = js['features']
    # return features
    return js


def write(gjson_object, geojson_filename):
    """
    Write GeoJSON object to file
    """
    with open(geojson_filename, 'w') as fp:
        json.dump(gjson_object, fp)
    return geojson_filename


def write_feature_json(feature:dict, reference_path_field:str='image_path') -> bool:
    """
    Write feature/metadata next to feature['image_path']
    """
    import json
    try:
        reference_path = feature['properties'][reference_path_field]
        json_filename = filenames.change_extension(reference_path, 'json')
        with open(json_filename, 'w') as fp:
            json.dump(feature, fp)
        return json_filename
    except Exception as err:
        raise err
        return None



import os
import sys
import shutil

import requests

from . import log

# === Inspect functions
# ref: https://fabianlee.org/2019/09/22/python-using-a-custom-decorator-to-inspect-function-arguments/
#
import functools
import inspect

def logintrospect(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        inspect_decorator(func, args, kwargs)
        result = func(*args, **kwargs)
        print("-return " + str(result))
        return result
    return wrapper


def inspect_decorator(func,args,kwargs):
  funcname = func.__name__
  print("function {}()".format(funcname))

  # get description of function params expected
  argspec = inspect.getargspec(func)

  # go through each position based argument
  counter = 0
  if argspec.args and type(argspec.args is list):
    for arg in args:
      # when you run past the formal positional arguments
      try:
        print("*" + str(argspec.args[counter]) + " = " + str(arg))
        counter+=1
      except IndexError as e:
        # then fallback to using the positional varargs name
        if argspec.varargs:
          varargsname = argspec.varargs
          print("*" + varargsname + " = " + str(arg))
        pass

  # finally show the named varargs
  if argspec.keywords:
    kwargsname = argspec.keywords
    for k,v in kwargs.items():
      print("**" + kwargsname + " " + k + " = " + str(v))

# ==========

@logintrospect
def _is_downloaded(url, filename):
    if not os.path.isfile(filename):
        log.debug("File '{}' does not exist.".format(filename))
        return False

    log.debug("File '{}' exist.".format(filename))
    try:
        r = requests.get(url, stream=True)
        remote_size = int(r.headers['Content-Length'])
        local_size = int(os.path.getsize(filename))
    except Exception as err:
        log.error(err)
        return False
    log.debug("Remote file size: " + str(remote_size))
    log.debug("Local file size: " + str(local_size))
    return local_size == remote_size


def download_file_silent(url, filename):
    """
    Download file (silently)

    Usage:
        download_file_silent('http://web4host.net/5MB.zip', 'local_filename.zip')
    """
    try:
        r = requests.get(url)
        with open(filename,'wb') as f:
            f.write(r.content)
    except Exception as err:
        log.error(err)
        return False

    return True


def download_file_progress(url, filename, verbose=False):
    """
    Download file with progressbar

    Usage:
        download_file_progress('http://web4host.net/5MB.zip', 'local_filename.zip')
    """
    import tqdm
    try:
        r = requests.get(url, stream=True)
        file_size = int(r.headers['Content-Length'])
        chunk = 1
        chunk_size=1024
        num_bars = int(file_size / chunk_size)
        if verbose:
            print(dict(file_size=file_size))
            print(dict(num_bars=num_bars))

        with open(filename, 'wb') as fp:
            for chunk in tqdm.tqdm(
                                        r.iter_content(chunk_size=chunk_size)
                                        , total= num_bars
                                        , unit = 'KB'
                                        , desc = filename
                                        , leave = True # progressbar stays
                                    ):
                fp.write(chunk)
    except Exception as err:
        log.error(err)
        return False

    return True


def download_file(url, filename=None, progress_on=False, make_dirs=True):
    if not filename:
        local_filename = os.path.join('.', url.split('/')[-1])
    else:
        local_filename = filename

    if _is_downloaded(url, local_filename):
        log.debug("File '{}' from '{}' already downloaded".format(local_filename, url))
        return local_filename

    _path = os.path.dirname(local_filename)
    if not os.path.isdir(_path):
        if make_dirs:
            os.makedirs(_path, exist_ok=True)
        else:
            print("Path '{}' does not exist.".format(_path))
            return None

    print('--> Downloading file {} ..'.format(local_filename))
    if progress_on:
        download_file_progress(url, local_filename)
    else:
        download_file_silent(url, local_filename)
    print('--> File downloaded.'.format(local_filename))

    return local_filename


def download_files(urls, descriptors, filenames=None, path=None, progress_on=False):
    assert isinstance(urls, (list, tuple))
    assert isinstance(descriptors, (list, tuple))
    assert len(urls) == len(descriptors)

    if path and path.strip():
        os.makedirs(path)
    else:
        path = '.'

    if not filenames:
        filenames = [os.path.join(path, url.split('/')[-1]) for url in urls]
    assert len(filenames) == len(urls), "List of 'filenames' must match length of 'urls'"

    files_downloaded = {}
    for url, descriptor, filename in zip(urls,descriptors, filenames):
        try:
            _fn = download_file(url, filename, progress_on)
            files_downloaded[descriptor] = _fn
        except Exception as err:
            print(err)
            print("If, moving next")

    return files_downloaded


file = download_file
files_list = download_files



"""
Query USGS/ODE API for image data products
"""
from npt import log


def ode(dataset: str, bbox: dict=None, product_id: str = None, match: str = 'intersect', bbox_ref:str='C0'):
    """
    Return GeoDataFrame with found data products as features

    Input:
    - dataset: name of the dataset (see `npt.datasets`)
    - bbox: (opt., see product_id) bounding-box to query overlapping products.
            Dictionary keys: minlat, maxlat, westlon, eastlon;
            Latitude/longitude values range: (-90:90, -180:180)
            If not defined, 'product-id' is expected.
    - product_id : (optional, see bbox) product-id in 'dataset'
            If not defined, 'bbox' is expected.
    - match: how to consider overlapping matching footprints.
             Options are: 'intersect', 'contain'
    - bbox_ref: if 'C0' (default), 'bbox' longitudes are centered at 0 (-180:180),
                if 'C180', 'bbox' longitudes are centered at 180 (0:360).
    """
    from npt.search._ode import ODE

    assert bbox or product_id, "Please, define either 'bbox' or 'product_id'"

    if bbox:
        prods = ODE(dataset).query_bbox(bbox=bbox, match=match, bbox_ref=bbox_ref).parse()
    if product_id:
        if bbox:
            # Not implemented yet: could filter for product-id(s) expression (exact or regex)
            pass
        prods = ODE(dataset).query_pid(product_id=product_id).parse()


    if not prods:
        return None

    return prods.to_dataframe()



import os
import shutil
import tempfile
from pathlib import Path

from npt import log

from ..utils.filenames import change_extension as _change_file_extension
from ..utils.filenames import change_dirname as _change_file_dirname
from ..utils.filenames import insert_preext as _add_file_subextension
from ..utils.geojson import write_feature_json

from npt.isis import format

def echo(msg):
    #print(msg)
    log.debug(msg)


def proj_planet2earth(filein, fileout):
    from npt.utils.raster import warp
    return warp(filein, fileout)


def from_geodataframe(gdf, dataset:str, basepath:str="./data/reduced/",
                 projection:str="sinusoidal",
                 tmpdir:str=None, keep_tmpdir:bool=False,
                 overwrite:bool=False):
    """
    Process all images in GeoDataFrame
    """
    import json
    gjson_obj = json.loads(gdf.to_json())

    new_gjson = from_geojson(gjson_obj, dataset, basepath,
                             projection,
                             tmpdir, keep_tmpdir,
                             overwrite)
    if not new_gjson:
        return None

    new_gdf = gdf.__class__.from_features(new_gjson['features'])
    return new_gdf

from_dataframe = from_geodataframe


def from_geojson(geojson:dict, dataset:str, basepath:str="./data/reduced/",
                 projection:str="sinusoidal",
                 tmpdir:str=None, keep_tmpdir:bool=False,
                 overwrite:bool=False):
    """
    Process all images in geojson features
    """
    from copy import copy as shallowcopy

    features = geojson['features']
    new_features = []
    for feature in features:
        new_feature = from_feature(feature, dataset, basepath,
                                    projection=projection,
                                    tmpdir=tmpdir,
                                    keep_tmpdir=keep_tmpdir,
                                    overwrite=overwrite)
        assert id(new_feature) != id(feature)
        # Write the image/product respective feature/metadata next to it
        feature_filename = write_feature_json(new_feature)
        print("Feature/metadata file '{}' written.".format(feature_filename))
        new_features.append(new_feature)

    new_geojson = shallowcopy(geojson)
    assert id(new_geojson) != id(geojson)
    new_geojson['features'] = new_features

    return new_geojson


def from_feature(geojson_feature, dataset:str, basepath:str="./data/reduced/",
                 projection:str="sinusoidal",
                 tmpdir:str=None, keep_tmpdir:bool=False,
                 overwrite:bool=False):
    echo("Processing Feature: {!s}".format(geojson_feature))
    echo("Output go to: {!s}".format(basepath))
    feature = geojson_feature.copy()
    properties = feature['properties']
    properties = _run_props(properties, dataset=dataset, basepath=basepath,
                            projection=projection,
                            tmpdir=tmpdir, keep_tmpdir=keep_tmpdir,
                            overwrite=overwrite)
    feature['properties'] = properties
    echo("Post-processed feature: {!s}".format(feature))
    return feature


def _run_props(properties:dict, dataset:str, basepath:str="./data/reduced/",
                 projection:str="sinusoidal",
                 tmpdir:str=None, keep_tmpdir:bool=False,
                 overwrite:bool=False):
    properties = properties.copy()
    image_filename = properties['image_path']
    tif = _run_file(image_filename,
        basepath=basepath,
        projection=projection,
        tmpdir=tmpdir,
        dataset=dataset,
        overwrite=overwrite,
        keep_tmpdir=keep_tmpdir)
    echo("Output file (TIF): {!s}".format(tif))
    if tif:
        # echo("IF 'out' (non-null output)")
        # assert len(out) == 2, "Was expecting a tuple of filenames (CUB,TIF). Instead got {!s}".format(out)
        # img_isis, img_tiff = out
        # echo("ISIS3/IMG output file: {!s}".format(img_isis))
        # echo("GeoTIFF output file: {!s}".format(img_tiff))
        #TODO: change 'image_path' to 'cube_path'. 'image_path' to stay pointing to source file
        try:
            tmp = tif.as_posix()
            tif = tmp
        except:
            assert isinstance(tif, str)
        # properties['image_path'] = tif
        properties['tiff_path'] = tif
        return properties
    else:
        log.error("Processing output is null. See the temp files.")
        return None


#TODO: Add argument to set docker container to run --e.g., isis3-- commands
def _run_file(filename:str, dataset:str, basepath:str="./data/reduced/",
                projection:str="sinusoidal",
                tmpdir:str=None, keep_tmpdir:bool=False,
                overwrite:bool=False,
                cog:bool=True, make_dirs:bool=True):
# def run_file(filename_init, output_path, map_projection="sinusoidal",
#             tmpdir=None, cog=True, make_dirs=True,
#             dataset=None, overwrite=True, keep_tmpdir=False):
    filename_out = Path(basepath) / f"{_change_file_extension(filename.split('/')[-1], 'tif')}"
    if not overwrite:
        if filename_out.exists():
            log.info(f"File '{filename_out}' already exists. Use overwrite to force rewriting.")
            return filename_out.as_posix()

    # Create a temp dir for the processing
    if tmpdir and not os.path.isdir(tmpdir):
        if make_dirs:
            os.makedirs(tmpdir, exist_ok=True)
        else:
            print("Path '{}' does not exist.".format(tmpdir))
            return None

    if tmpdir:
        assert os.path.isdir(tmpdir), """Given tmpdir '{}' does not exist""".format(tmpdir)
        tempfile.tempdir = tmpdir

    if basepath and not os.path.isdir(basepath):
        if make_dirs:
            os.makedirs(basepath, exist_ok=True)
        else:
            print("Path '{}' does not exist.".format(basepath))
            return None

    assert os.path.isdir(basepath), """Given output_path '{}' does not exist""".format(basepath)

    try:
        tmpdir = tempfile.mkdtemp(prefix='neanias_')
    except:
        log.error("Temporary directory ('{}') could not be created.".format(tmpdir))
        raise err
    else:
        log.info("Temp dir: '{}'".format(tmpdir))

    if 'ctx' in dataset:
        f_tif = reduce_ctx(filename, projection, tmpdir)
    elif 'hirise' in dataset:
        f_tif = reduce_hirise(filename, tmpdir)
    elif 'hrsc' in dataset:
        f_tif = reduce_hrsc(filename, tmpdir)
    else:
        raise NotImplementedError

    # f_tif_out = _change_file_extension(f_in, 'tif')
    # f_tif_out = _change_file_dirname(f_tif_out, output_path)
    # f_cub_out = _change_file_dirname(f_cub, output_path)
    # f_tif_out = _change_file_dirname(f_tif, output_path)
    f_tif_out = filename_out

    try:
        log.info("Copying from temp to archive/output path")
        shutil.move(f_tif, f_tif_out)
        # shutil.move(f_cub, f_cub_out)
    except Exception as err:
        # log.error("File '{}','{}' could not be created".format(f_cub_out, f_tif_out))
        log.error("File '{}' could not be created".format(f_tif_out))
        log.error("Temporary files, '{}' will remain. Remove them manually.".format(tmpdir))
        raise err
    finally:
        if not keep_tmpdir:
            log.info("CLEANING temporary files/directory ({})".format(tmpdir))
            shutil.rmtree(tmpdir)
        else:
            log.info("KEEPING temporary files/directory ({})".format(tmpdir))

    return f_tif_out


def reduce_ctx(filename_init, map_projection, tmpdir):
    try:
        f_in = shutil.copy(filename_init, tmpdir)
        log.info("File '{}' copied".format(filename_init))

        # FORMAT (pds->isis)
        # -- Transfrom PDS (IMG) into ISIS (CUB) file
        f_cub = _change_file_extension(f_in, 'cub')

        format.pds2isis(f_in, f_cub)
        # -- Init SPICE kernel
        # format.init_spice(f_cub)

        #TODO: be able to define at function level the container to run
        #       Functions in FORMAT and CALIBRATION, for example, could use a
        #       container 'isis3', whereas the FORMATting to TIFF, could use a
        #       different container, 'gispy' with "gdal" in it (not in 'isis3')

        # CALIBRATION
        from npt.isis import calibration
        f_cal = _add_file_subextension(f_cub, 'cal')
        calibration.radiometry(f_cub, f_cal)

        # -- Init SPICE kernel
        format.init_spice(f_cal)

        # MAP-PROJECTION
        from npt.isis import projection
        ## Define projection
        f_map = _add_file_subextension(f_cal, 'map')
        _flist = [f_cal]
        proj_file = projection.define_projection(_flist, projection=map_projection, tmpdir=tmpdir)
        ## Project
        projection.map_project(f_cal, f_map, proj_file)

        # FORMAT to TIFF
        f_tif = _change_file_extension(f_in, 'tif')
        format.isis2tiff(f_map, f_tif, cog=True)

    except Exception as err:
        log.error(err)
        raise err
    else:
        log.info("Processing finished, file '{}' created.".format(f_tif))

    return f_tif


def reduce_hirise(filename_init, tmpdir):
    lbl_file = '.'.join(filename_init.split('.')[:-1] + ['LBL'])
    assert os.path.exists(lbl_file), f"Expecting file {lbl_file}"

    try:
        from npt.utils import raster
        f_lbl = shutil.copy(lbl_file, tmpdir)
        f_jp2 = shutil.copy(filename_init, tmpdir)
        f_tif = _change_file_extension(f_lbl, 'tif')
        raster.lbl2cog(f_lbl, f_tif)

    except Exception as err:
        log.error(err)
        raise err

    else:
        log.info("Processing finished, file '{}' created.".format(f_tif))

    return f_tif


def reduce_hrsc(filename_init, tmpdir):
    try:
        # FORMAT (pds->isis)
        from npt.isis import format
        f_jp2 = shutil.copy(filename_init, tmpdir)
        # FORMAT to TIFF
        f_tif = _change_file_extension(f_jp2, 'tif')
        format.jpeg2tiff(f_jp2, f_tif, cog=True)

    except Exception as err:
        log.error(err)
        raise err

    else:
        log.info("Processing finished, file '{}' created.".format(f_tif))

    return f_tif



import sqlite3
from sqlite3 import Error

from npt import log


class DB(object):
    _conn = None
    def __init__(self, db_file='/tmp/npt.db'):
        super().__init__()
        # self.create_connection(db_file=':memory:')
        self.create_connection(db_file)

    def create_connection(self, db_file):
        self._conn = create_connection(db_file)

    def create_table(self, sql_create):
        create_table(self._conn, sql_create)

    def insert_record(self, sql_insert):
        insert_record(self._conn, sql_insert)



def create_connection(db_file):
    conn = None
    try:
        conn = sqlite3.connect(db_file)
        log.info("Connection {!s} established.".format(conn))
    except Error as e:
        log.error(e)
        raise e
    return conn


def create_table(conn, sql_create):
    c = _exec(conn, sql_create)


def insert_record(conn, sql_insert):
    cur = _exec(conn, sql_insert)
    if cur:
        conn.commit()
    return cur.lastrowid


def _exec(conn, sql):
    try:
        c = conn.cursor()
        c.execute(sql)
    except Error as e:
        log.error(e)
        raise e
    return c



db = DB()



"""
Provides functions to run `ISIS3`_ commands

Modules you'll find here handle tasks primarily thought to handle processing for
`NEANIAS`_ and `GMAP`_, where we are particularly interest on images from
Mars, the Moon, and Mercury.
Modules (``format``, ``calibration``, etc) provide some functions to process
and handle our (`PDS`_) data.

The `isissh` callable you'll find here is an interface to `sh`_ adapted to
the way ISIS3 define their command line interface (without hifen (``--``), with ``=``);
E.g,
.. code-block:: bash

   $ automos FROMLIST=mapcubs.list MOSAIC=mosaic.cub

E.g, :command: `automos FROMLIST=mapcubs.list mosaic=B01_009_mosaic.cub`

With ``isissh`` the command above would be called (here, in python):
.. code-block:: python
   isissh.automos(FROMLIST='mapcubs.list', MOSAIC=mosaic.cub)

**Always** use *UPPER-CASE* on those calls' arguments, that will free yourself
from any Python `reserved keywords`_ such as ``from``; You *cannot* call:
:command: isissh.spiceinit(from=filename)
, Python will give you a syntax error. But you *can* call:
.. code-block:: python
   isissh.spiceinit(FROM=filename)
 since ISIS3 command arguments are case *in*\ sensitive.

.. _ISIS3: bla
.. _NEANIAS: bla
.. _GMAP: bla
.. _sh: bla
"""
# import sh
# isissh = sh(_long_prefix="")

# from ._sh import sh
# isissh = sh    # Deprecated, here only for backwards compatibility

from ._sh import Sh as _Sh
sh = _Sh()
def set_docker(*args, **kwargs):
    sh.set_docker(*args, **kwargs)

# from . import (format,
#                calibration,
#                projection)


__all__ = ['isissh']



import os
# import rasterio
# import rioxarray
# from rioxarray import merge as rxmerge

from . import log


def copy_mean(merged_data, new_data, merged_mask, new_mask, **kwargs):
    return copy_(merged_data, new_data, merged_mask, new_mask,
        mean=True, **kwargs)


def copy_blend(merged_data, new_data, merged_mask, new_mask, **kwargs):
    return copy_(merged_data, new_data, merged_mask, new_mask,
        blend=True, **kwargs)


def copy_diff(merged_data, new_data, merged_mask, new_mask, **kwargs):
    return copy_(merged_data, new_data, merged_mask, new_mask,
        diff=True, **kwargs)


def copy_all(merged_data, new_data, merged_mask, new_mask, **kwargs):
    return copy_(merged_data, new_data, merged_mask, new_mask,
        diff=True, blend=True, mean=True, **kwargs)


def copy_(merged_data, new_data, merged_mask, new_mask,
          diff:bool=False, mean:bool=False, blend:bool=False,
          **kwargs):
    """
    Modify 'merged_data'
    """
    import numpy as np

    mask = np.empty_like(merged_mask, dtype="bool")
    np.logical_or(merged_mask, new_mask, out=mask)
    np.logical_not(mask, out=mask)

    if diff and mask.sum():
        # If 'diff', modify 'new_data' too!
        new_data = _diff(merged_data, new_data, mask)

    if mean:
        _mean(merged_data, new_data, mask)

    if blend:
        _blend(merged_data, new_data, mask)

    np.logical_not(new_mask, out=mask)
    np.logical_and(merged_mask, mask, out=mask)
    np.copyto(merged_data, new_data, where=mask, casting="unsafe")


def _blend(merged_data, new_data, mask, slope:str='gaussian', length:int=None):
    """
    Modify 'merged_data' by "blending" it to 'new_data' on difference slopes

    Input:
        slope: options are 'gaussian', 'uniform'
        length: size (in pixels) to smooth the slope (through the 'mask')
    """
    from scipy.ndimage import gaussian_filter, uniform_filter

    if slope == 'gaussian':
        _filter = gaussian_filter
    else:
        _filter = uniform_filter

    if not length:
        _size = min(mask.shape[1:])//3 # Make a large filter
    else:
        assert length > 1
        _size = int(length)

    weights = _filter(mask.astype(float), _size, mode='constant')
    weights = (weights - weights.min())/(weights.max() - weights.min())
    merged_data[mask] = (merged_data * (1 - weights) + new_data * weights)[mask]


def _mean(merged_data, new_data, mask):
    """
    Modify 'merged_data' with the average between merged_/new_data
    """
    import numpy as np

    _stack = np.stack([merged_data, new_data])
    merged_data[mask] = np.mean(_stack, axis=0)[mask]


def _diff(merged_data, new_data, mask):
    """
    Return 'new_data' + average difference between merged_/new_data
    """
    n_merged_data = merged_data[mask]
    n_new_data = new_data[mask]
    _mean = (n_merged_data - n_new_data).mean()
    return new_data + _mean


MERGE_METHODS = {
    'mean': copy_mean,
    'diff': copy_diff,
    'blend': copy_blend,
    'all': copy_all,
    'fast': copy_
}


def merge(filenames, output, nodata=None, method:str='all'):
    """
    Input:
        method: str
            How to resolve overlapping pixes.
            Options are: 'mean' and Rasterio's ('first,last, min, max')
            https://github.com/rasterio/rasterio/blob/master/rasterio/merge.py
    """
    from rioxarray import merge as rxmerge
    
    log.debug(f"Running '{merge.__name__}', from {__name__}")

    elements = []
    for file_ in filenames[::-1]:
        elements.append(rioxarray.open_rasterio(file_))

    method = MERGE_METHODS.get(method, method)
    merged = rxmerge.merge_arrays(elements, nodata=nodata, method=method)
    merged.rio.to_raster(output)
    return output



from npt import log

from . import _docker as docker
# try:
#     from . import _docker as docker
# except:
#     docker = None


class Sh(object):
    """
    Class meant to abstract the underlying Shell/OS layer

    It is similar to the package 'sh' (https://pypi.org/project/sh/) in the
    sense that it "wraps" command-line/shell commands to work seemlessly from
    inside Python.
    Different from pure 'sh' (which this class is based on),
    this class necessarily adds a routing layer: instead of doing `sh.ls('-l')`
    as you would with 'sh', here you have to do `ls = sh.wrap('ls'); ls('-l')`.
    Which *is similar* to 'sh.bare' method, but here is mandatory.

    There are three reasons that justify it:
    * commands are *always* run in a bash login shell,
    * docker containers can be set to seemlessly run commands,
    * a logging handler is set to *always* log the wrapped commands

    In case a (docker) container is used, and volumes are being used for I/O,
    this wrapper also *translates* host URL to corresponding ones inside docker.
    """
    _sh = None
    _maps = None

    def __init__(self):
        self.reset()

    def __call__(self, *args, **kwargs):
        log.debug(args)
        log.debug(kwargs)
        if self._maps:
            args = _map_args(args, self._maps)
            kwargs = _map_kwargs(kwargs, self._maps)
        return self._sh(*args, **kwargs)

    def reset(self):
        """
        Simply start a brand new shell.
        """
        _sh = _set_sh()
        log.debug(_sh)
        self._sh = _sh
        self._maps = None

    @staticmethod
    def log(res):
        try:
            msg = "Exit code: "+str(res and res.exit_code)
        except:
            msg = str(res)
        log.debug(msg)

    def set_docker(self, name, mappings=None, inspect=False):
        """
        Set running container 'name' to handle exec/commands

        Input:
            name: string
                Name of a running container
            mappings: list, dictionary
                Either a list/tuple mapping host to container paths,
                `` ('path_in_host', 'path_in_container') ``
                or a dictionary where keys are labels/args and values the tuples:
                `` { 'arg': ('path_in_host', 'path_in_container') } ``
        """
        if docker:
            assert name in docker.list_containers()
            self._sh = docker.bake_container(name)
            self._maps = mappings
        else:
            self.log("Docker is not defined.")

    def wrap(self, exec):
        """
        Return a callable wrapping command 'exec'

        This callable (really just a closure around Bash) accepts *any*
        argument(s) or kw-argument(s) you feel like running 'exec' with.

        Input:
            * exec : str
                Command name to wrap (e.g, "echo")
        """
        return _wrap(exec, sh_local=self)


def _set_sh():
    """
    Return a Bash login shell
    """
    from sh import bash
    return bash.bake('--login -c'.split())


def _map_args(args, map_paths):
    """
    Translate 'args' according to 'map_paths'

    'args' is like:
    ```
    ['dummy-arg', '/Volumes/at_host/bla.txt']
    ```
    And 'map_paths' (given at 'Sh.set_docker()') is like:
    ```
    { 'infile': ('/Volumes/at_host', '/mnt/at_container') }
    ```.
    """
    log.debug(args)
    log.debug(map_paths)

    map_tuples = map_paths.values() if isinstance(map_paths, dict) else map_paths

    _args = []
    for val in args:
        for _host, _cont in map_tuples:
            _val = val.replace(_host, _cont)
            if _val != val:
                break
        _args.append(_val)

    # maps = map_paths
    # _args = []
    # # for arg in args:
    # #     for _host,_cont in map_paths.values():
    # #         arg = arg.replace(_host, _cont)
    # #     _args.append(arg)
    # # return _args
    # if isinstance(maps, dict):
    #     for val in args:
    #         try:
    #             # Try for the mapping "{ 'infile': ('/Volumes/at_host', '/mnt/at_container') }"
    #             _host = maps[key][0]
    #             _cont = maps[key][1]
    #         except Exception as err:
    #             log.error(err)
    #             # If not -- either because 'key' doesn't exist or value is not a tuple --
    #             # just pass key:val forward.
    #             _val = val
    #         else:
    #             # Translate path from host URL to container's reference.
    #             _val = val.replace(_host, _cont)
    #
    #         _args.append(_val)
    #
    # elif isinstance(maps, (list,tuple)):
    #     for val in args:
    #         try:
    #             _host, _cont = maps
    #         except Exception as err:
    #             log.error(err)
    #             _val = val
    #         else:
    #             # Translate path from host URL to container's reference.
    #             _val = val.replace(_host, _cont)
    #
    #         _args.append(_val)
    #
    # else:
    #     # If there is no map defined (for some fucking reason), just map trivially
    #     # _kw.update(kwargs)
    #     assert False, "Mapping paths should be defined. This line should never be hit!"

    return _args


def _map_kwargs(kwargs, map_paths):
    """
    Translate 'kwargs' values according to 'map_paths' mappings

    'kwargs' is something like:
    ```
    { 'infile' : '/Volumes/at_host/bla.txt' }
    ```.

    And 'map_paths' (given at 'Sh.set_docker()') is like:
    # ```
    # TODO: implement this format!
    # { '/Volumes/at_host' : '/mnt/at_container' }
    # ```
    # or
    ```
    { 'infile': ('/Volumes/at_host', '/mnt/at_container') }
    ```.
    The second form is used only when you want to restrict such mapping/translation
    to a specific command-line argument
    """
    log.debug(kwargs)
    log.debug(map_paths)

    map_tuples = map_paths.values() if isinstance(map_paths, dict) else map_paths

    _kw = {}
    for key, val in kwargs.items():
        for _host, _cont in map_tuples:
            _val = val.replace(_host, _cont)
            if _val != val:
                break
        _kw[key] = _val

    # maps = map_paths
    # _kw = {}
    # if isinstance(maps, dict):
    #     for key, val in kwargs.items():
    #         try:
    #             # Try for the mapping "{ 'infile': ('/Volumes/at_host', '/mnt/at_container') }"
    #             _host = maps[key][0]
    #             _cont = maps[key][1]
    #         except Exception as err:
    #             log.error(err)
    #             # If not -- either because 'key' doesn't exist or value is not a tuple --
    #             # just pass key:val forward.
    #             _val = val
    #         else:
    #             # Translate path from host URL to container's reference.
    #             _val = val.replace(_host, _cont)
    #
    #         _kw[key] = _val
    #
    # elif isinstance(maps, (list,tuple)):
    #     for key, val in kwargs.items():
    #         try:
    #             _host, _cont = maps
    #         except Exception as err:
    #             log.error(err)
    #             _val = val
    #         else:
    #             # Translate path from host URL to container's reference.
    #             _val = val.replace(_host, _cont)
    #
    #         _kw[key] = _val
    #
    # else:
    #     # If there is no map defined (for some fucking reason), just map trivially
    #     # _kw.update(kwargs)
    #     assert False, "Mapping paths should be defined. This line should never be hit!"

    return _kw


def _wrap(exec, sh_local, log=log):
    """
    Return 'sh' to be called with arguments to 'exec'

    The wrapped 'sh' reports to log.debug and parse the (future) arguments
    to build a string for sh('exec <future-args>').

    Input:
        * exec: string
            command to execute by the container
        * sh_local: Sh
        * log: logging-handler

    Output:
        Return callable/function closure with 'exec' wrapped
    """
    if isinstance(exec, str):
        exec = [exec]

    def _sh(*args,**kwargs):
        """
        Run and return result of 'exec' in 'sh_local' with argument 'args/kwargs'
        """
        v = [f'{v}' for v in args]
        kv = [f'{k}={v}' for k,v in kwargs.items()]
        comm = ' '.join(exec+v+kv)
        log.debug(comm)
        # 'comm' is effectively the full/command-line to run
        return sh_local(comm)

    return _sh


# # Global/Singleton
# sh = Sh()



import json

from npt import log

from . import tmpdir


def read_geojson(filename):
    """
    Return JSON object from GeoJSON
    """
    with open(filename, 'r') as fp:
        js = json.load(fp)
    return js



"""
Define commands and sub-commands line interface
"""
import typer

from npt import pipelines

app = typer.Typer()


@app.command()
def datasets():
    """List the supported datasets"""
    from npt import datasets
    for dset in datasets.list():
        typer.echo(dset)


@app.command()
def search(dataset:str, bbox:str, match:str='intersect', bbox_ref:str='C0',
            output_geojson:str=''):
    """
    Write GeoJSON with products the bounding-box intersects/contains

    Input:
        dataset: name of the dataset (see `datasets`)
            Ex: 'mars/mro/ctx/edr', 'mars/mex/hrsc/refdr3', 'mars/mro/hirise/rdrv11'
        bbox: bounding-box to query overlapping products.
            Dictionary keys: minlat, maxlat, westlon, eastlon;
            Latitude/longitude values range: (-90:90, -180:180)
        match: how to consider overlapping matching footprints.
            Options are: 'intersect', 'contain'
        bbox_ref: which meridian to consider as world/coordinates center ('C0' or 'C180')
            If 'C0' (default), consider longitude range as [-180:180].
            Use 'C180' to consider (longitude) coordinates in the range [0:360].
        output_geojson: path/filename for the output GeoJSON file.
    """
    from npt import pipelines

    if not bbox.strip():
        typer.echo("No bounding-box defined")

    output = output_geojson if output_geojson.strip() else None

    res = pipelines.search(dataset=dataset, bbox=bbox, match=match,
                            bbox_ref=bbox_ref, output_geojson=output)

    typer.echo(res)


if __name__ == '__main__':
    app()

# import sys
# import click
# from click import argument,option
#
# from npt import log
# from npt import datasets
# from npt.search import ode
# from npt.download import from_geojson as download_geojson
# from npt.reduce import from_geojson as reduce_geojson
#
# from npt.utils import geojson
# from npt.utils.bbox import string_2_dict as bbox_string_2_dict
# from npt.utils.formatters import json_2_geojson
#
#
# # === MAIN
# @click.group()
# @option('--debug', is_flag=True, default=False, help="Output DEBUG-level messages")
# def main(debug:bool):
#     """
#     Interface to npt pipelines
#     """
#     if debug:
#         log.setLevel('DEBUG')
# # ===
#
#
# @main.command()
# def datasets_list():
#     """List the supported datasets"""
#     for dset in datasets.list():
#         print(dset)
#
#
# @main.command()
# @argument('dataset')
# @argument('bbox')
# @argument('output_geojson')
# @option('--contains/--intersects', is_flag=True, default=False, help="Bounding-box intersects or contains products' footprint")
# @option('--coordsref', default='C0', help="Central coordinate reference: 'C0'(default) or 'C180'.")
# def search(dataset:str, bbox:str, output_geojson:str, contains:bool, coordsref:str):
#     """
#     Query 'provider'/'dataset' for data products in/on 'bbox'
#
#     \b
#     Attributes:
#         dataset: See `npt datasets-list` for options
#         bbox: bounding-box definition, '[minlat,maxlat,westlon,eastlon]'
#             Eg: if coords-ref=='C180': "[-0.5,0.5,359.5,0.5]"
#             Eg: if coords-ref=='C0'    "[-0.5,0.5,-0.5,0.5]"
#         output_geojson: output GeoJSON filename
#         contains: if True, consider only footprints inside 'bbox' (False: all intersecting)
#     """
#     assert len(output_geojson.strip())>0
#
#     bbox = bbox.replace('[','').replace(']','')
#     bounding_box = bbox_string_2_dict(bbox)
#     log.debug(bounding_box)
#
#     match = 'contain' if contains else 'intersect'
#
#     products = ode(bbox=bounding_box, dataset=dataset, match=match, bbox_ref=coordsref)
#
#     if products is None:
#         print("No products found.", file=sys.stderr)
#         return None
#     else:
#         products.to_file(output_geojson, driver='GeoJSON', index=False)
#         return output_geojson
#
#
# @main.command()
# @argument('input_geojson')
# @argument('dest_path')
# @argument('output_geojson')
# @option('--progress/--silent', is_flag=True, default=True, help="Print download progress")
# def download(input_geojson:str, dest_path:str, output_geojson:str, progress:bool):
#     """
#     Download features' image_url/label_url data products
#     """
#     assert len(output_geojson.strip())>0
#
#     input_gjson = geojson.read(input_geojson)
#     log.debug(input_gjson)
#
#     result_gjson = download_geojson(input_gjson, basepath=dest_path, progressbar=progress)
#     log.debug(result_gjson)
#
#     geojson.write(result_gjson, output_geojson)
#
#     return output_geojson
#
#
# @main.command()
# @argument('input_geojson')
# @argument('dataset')
# @argument('dest_path')
# @argument('output_geojson')
# @option('--projection', default='sinusoidal', help="Projection to lay images.")
# @option('--tmpdir', default='', help="Temp-dir to use. Empty/none means a random one will be used.")
# @option('--keep-tmpdir', is_flag=True, default=False, help="Keep temp-dir (don't delete it) if things fail.")
# @option('--overwrite', is_flag=True, default=False, help="Force overwriting of products.")
# @option('--docker-isis', default='', help="Container name to use for processing.")
# def reduce(input_geojson:str, dataset:str, dest_path:str, output_geojson:str,
#                  projection:str,
#                  tmpdir:str, keep_tmpdir:bool,
#                  overwrite:bool,
#                  docker_isis:str):
#     """
#     Reduce image data to analysis/publication-ready (GeoTiff)
#     """
#     assert len(output_geojson.strip())>0
#
#     input_gjson = geojson.read(input_geojson)
#     log.debug(input_gjson)
#
#     if docker_isis:
#         from npt import isis
#         isis.set_docker(docker_isis)
#
#     result_gjson = reduce_geojson(input_gjson, dataset=dataset,
#                                   basepath=dest_path, tmpdir=tmpdir)
#     log.debug(result_gjson)
#
#     geojson.write(result_gjson, output_geojson)
#
#     return output_geojson
#
#
#
# @main.command()
# @argument('geojson_file')
# @argument('basepath')
# @option('--output', metavar='<.geojson>', default='', help="GeoJSON filename with query results")
# @option('--tmpdir', default=None, help="Temp dir to use during processing")
# def mosaic(geojson_file, basepath, output, tmpdir):
#     """
#     Make mosaic from files in 'input_geojson' file. Write GeoJSON with mosaic feature.
#     """
#     return Mosaic.run(geojson_file, basepath, output, tmpdir)
#
#
# if __name__ == '__main__':
#     main()



"""
Module to interface MEEO pipelines manager
"""
from npt import log

def search(dataset, bbox, match='intersect', bbox_ref='C0', output_geojson=None):
    """
    Write GeoJSON with products the bounding-box intersects/contains

    Input:
        dataset: name of the dataset (see `npt.datasets`)
            Ex: 'mars/mro/ctx/edr', 'mars/mex/hrsc/refdr3', 'mars/mro/hirise/rdrv11'
        bbox: bounding-box to query overlapping products.
            Dictionary keys: minlat, maxlat, westlon, eastlon;
            Latitude/longitude values range: (-90:90, -180:180)
        match: how to consider overlapping matching footprints.
            Options are: 'intersect', 'contain'
        bbox_ref: which meridian to consider as world/coordinates center ('C0' or 'C180')
            If 'C0' (default), consider longitude range as [-180:180].
            Use 'C180' to consider (longitude) coordinates in the range [0:360].
        output_geojson: path/filename for the output GeoJSON file.
    """
    from npt.search import ode

    if not bbox:
        return None

    gdf = ode(dataset=dataset, bbox=bbox, match=match, bbox_ref=bbox_ref)

    if gdf is None:
        log.info('Search returned no results.')
        return None

    if output_geojson:
        gdf.to_file(output_geojson, driver='GeoJSON', index=False)
        return output_geojson
    else:
        return gdf


def download(feature, basepath, progressbar=False):
    """
    Download data products and return updated feature (with 'image_path'/'label_path' added)

    Input:
    * feature:
        GeoJSON feature (from ODE search)
        Property 'image_url' is expected, field 'image_path' will be added
        Property 'label_url' is optional, field 'label_path' will be added
    * basepath:
        Base filesystem path (directory) where product will be downloaded
    * progressbar (False):
        If a (tqdm) progress-bar should show download progress
    """
    from npt.download import from_feature

    new_feature = from_feature(feature, basepath, progressbar=progressbar)

    return new_feature


def reduce(feature, basepath, tmpdir=None, keep_tmpdir=True, overwrite=False):
    """
    Reduce image to "science-ready" level, return updated feature

    Note: property 'tiff_path' is added to (new) feature

    Input:
    * feature:
        GeoJSON feature, with 'image_path' property
    * basepath:
        Base filesystem path (directory) where product will be downloaded
    * tmpdir:
        Temp directory where to do the processing/intermediate steps
    * keep_tmpdir:
        When and error occurs, to keep (True) or not to keep (False) temp data
    * overwrite:
        If False and output (file) already exists, skip processing
    """
    from npt.reduce import from_feature

    projection="sinusoidal"

    props = feature['properties']
    dataset = '/'.join(['mars', props['mission'], props['inst'], props['type']]).lower()

    new_feature = from_feature(feature, dataset, basepath=basepath, projection=projection,
                                tmpdir=tmpdir, keep_tmpdir=keep_tmpdir, overwrite=overwrite)

    return new_feature


def mosaic(geojson, basepath:str, output_geojson=None, method='all', scale_factor=0.1):
    """
    Create Mosaic from images in 'geojson' input, return geojson with one mosaic/feature
    """
    from npt.mosaic import mosaic

    gdf = mosaic(geojson, basepath=basepath, method=method, scale_factor=scale_factor)

    if gdf is None:
        log.info('Mosaic failed to create.')
        return None

    if output_geojson:
        gdf.to_file(output_geojson, driver='GeoJSON', index=False)
        return output_geojson
    else:
        return gdf



import os
from npt import log
from . import sh


def define_projection(list_files, projection='sinusoidal', precision=0, tmpdir=None):
    tmpdir = tmpdir or ''
    _file_list = os.path.join(tmpdir, 'files_to_project.list')
    with open(_file_list, 'w') as fp:
        fp.write('\n'.join(list_files))
        fp.write('\n')
    _file_proj = os.path.join(tmpdir, projection+'.map')

    mosrange = sh.wrap('mosrange')
    res = mosrange(FROMLIST=_file_list, TO=_file_proj,
                   PROJECTION=projection, PRECISION=precision)
    sh.log(res)
    return _file_proj


def map_project(filename_in, filename_out, filename_proj):
    cam2map = sh.wrap('cam2map')
    res = cam2map(FROM=filename_in, TO=filename_out, MAP=filename_proj, PIXRES='map')
    sh.log(res)
    return res



import os
# import osgeo

from . import log

COG_ARGS = (
    '-of COG',
    '-co BLOCKSIZE=512',
    '-co COMPRESS=LZW',
    '-co NUM_THREADS=2',
    '-co GEOTIFF_VERSION=AUTO'
    )

def warp(filenames, output):
    """
    Return filename of merged 'filenames' GeoTIFFs

    Input:
        filenames : list
            List of filenames to merge
        output : string
            Mosaic filename
    """
    from osgeo import gdal
    log.debug("Running 'warp' method.")

    # The merge function returns a single array and the affine transform info
    gdal.Warp(output, filenames, format="GTiff",
              options=["COMPRESS=LZW", "TILED=YES"])

    return output


def tif2cog(filein, fileout):
    """
    Transform GeoTIFF in COG
    """
    from osgeo import gdal

    cog_args = ' '.join(COG_ARGS)
    # print(cog_args)
    gdal.Translate(fileout, filein, options=cog_args)

    return fileout

tiff2cog = tif2cog


def lbl2cog(filein, fileout):
    """
    Transform the respective file from LABEL in filein to COG
    """
    print(filein, fileout)
    return tif2cog(filein, fileout)


def virtual(filenames, output):
    """
    Makes a mosaic through GDAL virtual file + translate
    """
    log.debug("Running 'virtual' method.")
    _froot, _ = os.path.splitext(output)
    _fvrt = _froot + ".vrt"

    vrt = gdal.BuildVRT(_fvrt, filenames, allowProjectionDifference=True)
    gdal.Translate(output, vrt)
    vrt = None

    return output


def merge(filenames, output):
    # from npt.isis import sh
    import subprocess
    log.debug("Running 'gdal-merge' method.")

    # gdal_merge = sh.wrap('gdal_merge.py')
    cog_args = ' '.join([
        # '-of COG',
        # '-co BLOCKSIZE=512',
        '-co COMPRESS=LZW',
        # '-co NUM_THREADS=2',
        # '-co GEOTIFF_VERSION=AUTO'
        ]).split()
    log.debug(cog_args)
    # res = gdal_merge('-o', output, *cog_args, *filenames)
    try:
        res = subprocess.call(['gdal_merge.py','-o',output] + cog_args + filenames)
    except Exception as err:
        log.error(str(err))
        return None
    else:
        return output



"""
Datasets dynamic import
"""
# Refs:
# - https://packaging.python.org/guides/creating-and-discovering-plugins/
# - https://stackoverflow.com/a/64124377/687896
import npt


def list():
    """
    Return (sorted) list of available datasets
    """
    # dsets = [d['id'] for d in _datasets]
    # return sorted(dsets)
    return sorted(_datasets.keys())


def descriptors(dataset_id):
    """
    Return descriptors set for given 'dataset'
    """
    return _datasets[dataset_id].get('descriptors')


def filters(dataset_id):
    """
    Return filters set for given 'dataset'
    """
    try:
        _filters = _datasets[dataset_id].get('filters')
    except:
        _filters = None
    return _filters

# -----------------------------------------
# Dynamic load module in current directory:
# (Not being used anymore, but for records;
#  It implements a kind of plugin for data)
#
_bla = '''
import sys
import pkgutil
import importlib

pkg = sys.modules[__package__]

_datasets = []
for finder, name, ispkg in pkgutil.iter_modules(pkg.__path__, pkg.__name__+'.'):
    mod = importlib.import_module(name)
    _datasets.append(mod)

del pkg, finder, name, ispkg
del importlib, pkgutil, sys
'''
# -----------------------------------------


## Descriptors are the fields/values of ODE Results the identify the data products
## of our interest per project. Here is template, each project have them different.
# _DESCRIPTORS_TEMPLATE = {
#     'product_image': ('Description', 'PRODUCT DATA FILE'),
#     'product_label': ('Description', 'PRODUCT LABEL FILE'),
#     'browse_image': ('Description', 'BROWSE IMAGE'),
#     'browse_thumbnail': ('Description', 'THUMBNAIL IMAGE')
# }

## Filters are (regex) applied over Product IDs (project-based).
## Filters are tuples: "(<regex>, <True=include, or False=exclude>)" from output
# FILTERS = {
#     'ctx': ("^(CRU|MOI|T01)_", False),
#     'hirise': ("^(PSP|ESP)_.*(RED)", True),
#     'hrsc': (".*_ND3.*", True)
# }


# FIXME: datasets should be an object "label:{sets}" (not a list/anonymous)
_datasets = [
    {
        "id": "{target}/{mission}/{instrument}/{product_type}",
        "target": "mars",
        "mission": "mro",
        "instrument": "ctx",
        "product_type": "edr",
        "descriptors": {
            'product_image': ('Description','PRODUCT DATA FILE WITH LABEL'),
        },
        "filters": {
            'product-id': ("^(CRU|MOI|T01)_", False)
        },
    },

    {
        "id": "{target}/{mission}/{instrument}/{product_type}",
        "target": "mars",
        "mission": "mro",
        "instrument": "hirise",
        "product_type": "rdrv11",
        "descriptors": {
            'product_image': ('Description', 'PRODUCT DATA FILE'),
            'product_label': ('Description', 'PRODUCT LABEL FILE'),
            'browse_image': ('Description', 'BROWSE'),
            'browse_thumbnail': ('Description', 'THUMBNAIL')
        },
        "filters": {
            'product-id': ("^(PSP|ESP)_.*(RED)", True)
        },
    },

    {
        "id": "{target}/{mission}/{instrument}/{product_type}",
        "target": "mars",
        "mission": "mex",
        "instrument": "hrsc",
        "product_type": "rdrv3",
        "descriptors": {
            'product_image': ('Description', 'PRODUCT DATA FILE'),
            'product_label': ('Description', 'PRODUCT LABEL FILE'),
            'browse_image': ('Description', 'BROWSE IMAGE'),
        },
        "filters": {
            'product-id': (".*_ND3.*", True)
        },
    },

    {
        "id": "{target}/{mission}/{instrument}/{product_type}",
        "target": "mars",
        "mission": "mex",
        "instrument": "hrsc",
        "product_type": "refdr3",
        "descriptors": {
            'product_image': ('Description', 'PRODUCT DATA FILE'),
            'product_label': ('Description', 'PRODUCT LABEL FILE'),
            'browse_image': ('Description', 'BROWSE IMAGE'),
        },
        "filters": {
            'product-id': (".*_ND3.*", True)
        },
    }
]


def _solve_datasets_references(dsets: list) -> dict:
    """
    Solve datasets/dictionary internal references


    """
    def _solve_self_references(mapin: dict) -> dict:
        """
        Return version of 'maps_in' where simple "{internal}-{reference}" are resolved.

        Example:
            input:
            ```
            {
                "id": "{target}/{mission}/{instrument}/{product_type}",
                "target": "some_body",
                "mission": "some_spacecraft",
                "instrument": "some_sensor",
                "product_type": "some_dataset"
            }
            ```
            output:
            ```
            {
                "id": "some_body/some_spacecraft/some_sensor/some_dataset",
                "target": "some_body",
                "mission": "some_spacecraft",
                "instrument": "some_sensor",
                "product_type": "some_datalevel"
            }
            ```
        """
        def _resolved(value):
            """
            Return True if 'value' has NO more "{}": all fields are resolved.
            """
            import re
            return not re.match('.*{.+}.*', value)

        from copy import deepcopy
        mappings = deepcopy(mapin)

        # Fields/Keys to skip dereferencing: we only map flate/self references,
        # values that are not "str" (ie, data-structures) are moved to 'map_out'.
        # 'map_aux' is composed by all values already "resolved" (without "{}")
        keys_bypass = [k for k,v in mappings.items() if not isinstance(v,str)]
        mapout = {k:mappings.pop(k) for k in keys_bypass}
        mapaux = {k:v for k,v in mappings.items()
                      if _resolved(v)}
        cnt = 0
        while len(mapaux) < len(mappings):
            # if 'mapaux' is smaller than 'mappings' it means non-resolved values still there
            cnt += 1
            assert cnt <= len(mappings), f"Apparently going for an infinite loop: {mappings} {mapaux}"
            _reset = set(mappings.keys()) - set(mapaux.keys())
            _aux = {k:mappings[k].format(**mapaux) for k in _reset}
            mapaux.update({ k:v for k,v in _aux.items()
                                if _resolved(v) })

        # Once 'mapaux' has same size as 'mappings' (all fields resolved),
        # update 'map-out' with flat-fields resolved.
        mapout.update(mapaux)
        return mapout

    _dsets = [_solve_self_references(d) for d in dsets]
    # Make the output a pivot table with dataset/objects for keyword "dataset 'id'"
    _pivot = {d['id']:d for d in _dsets}
    return _pivot

# Overwrite '_datasets' with parsed and pivot (list to dict{'id',...}) version
_datasets = _solve_datasets_references(_datasets)



import os
import tempfile
import pathlib

from npt import log

def mkpath(basedir=None, prefix=None):
    """
    Return path of directory for temporary use

    Args:
        basedir: string
            Base path (already existent) for parenting new temp-dir
        prefix: string
            Prefix to use for new temp-dir name (a random string is appended to it)

    Return:
        Path (pathlib) object for newly created temp-dir
    """

    if basedir:
        assert os.path.isdir(basedir), "Given basedir '{}' does not exist".format(basedir)
        tempfile.tempdir = basedir

    try:
        tmpdir = tempfile.mkdtemp(prefix=prefix)
    except:
        log.error("Temporary directory ('{}') could not be created.".format(tmpdir))
        raise err
    else:
        log.info("Temp dir: '{}'".format(tmpdir))

    return pathlib.Path(tmpdir)



import logging
# fmt="%(levelname)s:%(funcName)s():%(lineno)i: %(message)s"
fmt="%(levelname)s:%(module)s.%(funcName)s(): %(message)s"
logging.basicConfig(level=logging.INFO, format=fmt)
log = logging.getLogger('npt')
log.setLevel('INFO')
log.set_level = log.setLevel



import os
from copy import deepcopy
from copy import copy as shallowcopy

from npt import log

from npt.utils.download import download_file


def from_geodataframe(gdf, basepath:str,
                 image_url_field:str='image_url',
                 label_url_field:str='label_url',
                 image_path_field:str='image_path',
                 label_path_field:str='label_path',
                 progressbar:bool=False):
    """
    Download all entries from GeoDataFrame indicated by URLs in the resp. paths
    """
    import json
    gjson_obj = json.loads(gdf.to_json())

    new_gjson = from_geojson(gjson_obj, basepath,
                            image_url_field,
                            label_url_field,
                            image_path_field,
                            label_path_field,
                            progressbar)

    if not new_gjson:
        return None

    new_gdf = gdf.__class__.from_features(new_gjson['features'])
    return new_gdf

from_dataframe = from_geodataframe


def from_geojson(geojson:dict, basepath:str,
                 image_url_field:str='image_url',
                 label_url_field:str='label_url',
                 image_path_field:str='image_path',
                 label_path_field:str='label_path',
                 progressbar:bool=False) -> dict:
    """
    Download all entries (features) in GeoJSON, URLs in the resp. paths given
    """
    features = geojson['features']
    new_features = []
    for feature in features:
        new_feature = from_feature(feature, basepath,
                                    image_url_field,
                                    label_url_field,
                                    image_path_field,
                                    label_path_field,
                                    progressbar)
        assert id(new_feature) != id(feature)
        new_features.append(new_feature)

    new_geojson = shallowcopy(geojson)
    assert id(new_geojson) != id(geojson)
    new_geojson['features'] = new_features

    return new_geojson


#TODO: Add argument "data product type" to define what to download.
def from_feature(feature:dict, basepath:str,
                 image_url_field:str='image_url',
                 label_url_field:str='label_url',
                 image_path_field:str='image_path',
                 label_path_field:str='label_path',
                 progressbar:bool=False):
    """
    Download data products (Image, Label) inside 'basepath'
    """
    from ._pds import split_label_from_image
    from ._pds import write_feature_json

    properties = deepcopy(feature['properties'])
    geometry = deepcopy(feature['geometry'])

    image_url = properties.pop(image_url_field)

    label_url = None
    if label_url_field in properties:
        label_url = properties.pop(label_url_field)

    if (label_url is not None and label_url != image_url):
        label_path = _download(label_url, basepath=basepath, progressbar=progressbar)
    else:
        label_path = split_label_from_image(image_path)

    properties[label_path_field] = label_path

    image_path = _download(image_url, basepath=basepath, progressbar=progressbar)
    properties[image_path_field] = image_path

    # Update metadata set basaed on image Label.
    # FIXME: these MEEO metadata mappings are a bit chaotic
    metaDict = _createLblDict(label_path)
    metaDict.update(_map_meta_meeo(properties))
    properties.update(metaDict)

    new_feature = shallowcopy(feature)
    new_feature['properties'] = properties
    new_feature['geometry'] = geometry

    # Write the image/product respective feature/metadata next to it
    feature_filename = write_feature_json(new_feature)
    print("Feature/metadata file '{}' written.".format(feature_filename))

    return new_feature


def _download(url, basepath, progressbar=False):
    _file = url.split('/')[-1]
    file_path = os.path.join(basepath, _file)
    #TODO: check if file_path is here already, if so, skip downloading
    try:
        _out = download_file(url, filename=file_path, progress_on=progressbar)
    except Exception as err:
        log.error(err)
        return None

    return file_path


def _createLblDict( filename ):
    """
    filename is a 'LBL' (PDS LBL) filename
    """
    text = open(filename, 'r').read()
    lines = text.splitlines()
    img_header = {}
    for line in lines:
        value = line
        fields = [v.strip() for v in value.split("=")]
        key_value = fields[0]
        fill_value = fields[-1].replace('"',"").rstrip()
        if 'DATA_SET_ID' == key_value:
            # l_datasetId = fill_value.replace("-","_").replace(".","_").split("_")
            # img_header['datasetId'] = "_".join(l_datasetId[0:3:2])
            # img_header['observationMode'] = l_datasetId[4]
            pass

        # elif 'PRODUCT_ID' == key_value:
        #     img_header['idFromProvider'] = fill_value

        elif 'INSTRUMENT_HOST_NAME' == key_value:
            img_header['instrumentHostName'] = fill_value
            # img_header['spacecraftId'] = "".join([s[0].upper() for s in fill_value.split()])

        elif 'INSTRUMENT_NAME' == key_value:
            img_header['instrumentName'] = fill_value

        # elif 'INSTRUMENT_ID' == key_value:
        #     img_header['instrumentId'] = fill_value

        # elif 'TARGET_NAME' == key_value:
        #     img_header['targetName'] = fill_value

        elif 'MISSION_PHASE_NAME' == key_value:
            img_header['missionPhaseName'] = fill_value

        # elif 'PRODUCT_CREATION_TIME' == key_value:
        #     img_header['productCreationTime'] = fill_value

        # elif 'START_TIME' == key_value:
        #     img_header['startTime'] = fill_value.replace(" ", "")

        # elif 'STOP_TIME' == key_value:
        #     img_header['stopTime'] = fill_value.replace(" ", "")


    # img_header['solarDistance'] = ""
    # img_header['solarLongitude'] = ""

    _keys = list(img_header.keys())
    _keys.sort()
    return {k:img_header[k] for k in _keys}

def _map_meta_meeo(props):
    meta = {}
    meta['idFromProvider'] = props['id']
    meta['observationMode'] = props['type']
    meta['instrumentId'] = props['inst']
    meta['spacecraftId'] = props['mission']
    meta['datasetId'] = f"{props['mission']}_{props['inst']}"
    meta['targetName'] = props['Target_name']
    meta['productCreationTime'] = props['Product_creation_time']
    meta['startTime'] = props['UTC_start_time']
    meta['stopTime'] = props['UTC_stop_time']
    return meta



import requests
from npt import log
from npt import datasets


DB_ID = 'usgs_ode'

API_URL = 'https://oderest.rsl.wustl.edu/live2'

COORDS_REF = {
    'C0': 'Footprint_C0_geometry',  # -180:+180
    'C180': 'Footprint_GL_geometry' # 0:360
}

METADATA = [
    'Target_name',
    'Footprints_cross_meridian',
#    'Map_resolution'
    'Map_scale',
    'Center_latitude',
    'Center_longitude',
    'Easternmost_longitude',
    'Westernmost_longitude',
    'Minimum_latitude',
    'Maximum_latitude',
    'Emission_angle',
    'Incidence_angle',
    'Phase_angle',
    'Solar_longitude',
    'Observation_time',
    'Product_creation_time',
    'UTC_start_time',
    'UTC_stop_time'
]


class ODEProducts(list):
    def __init__(self):
        super()

    def to_dataframe(self):
        import shapely
        import geopandas
        from copy import deepcopy

        products = []
        for prod in self:
            product = deepcopy(prod)
            try:
                _geom = shapely.wkt.loads(product['geometry'])
                if (type(_geom) == shapely.geometry.GeometryCollection
                    or type(_geom) == shapely.geometry.MultiPolygon):
                    _geom = _geom.envelope
                assert type(_geom) == shapely.geometry.Polygon
            except Exception as err:
                print("Error in: ", product)
                _geom = None
            finally:
                product['geometry'] = _geom
            products.append(product)

        gdf = geopandas.GeoDataFrame(products)
        return gdf


class ODE(object):
    """
    Handles querying ODE, and then parsing the results for a given dataset

    Check 'npt.datasets.list' for the available/supported datasets
    """
    _result = None # Cache ODE query results
    _ref_coords = 'C0'

    def __init__(self, dataset):
        """
        dataset string example: 'mars/mro/ctx/edr'
        """
        target, mission, instrument, product_type = dataset.split('/')
        assert dataset in datasets.list(), f"Dataset '{dataset}' not recognised."
        self.dataset = dataset
        self.target = target
        self.mission = mission
        self.instr= instrument
        self.ptype = product_type

    def query_bbox(self, bbox, match='intersect', bbox_ref='C0'):
        """
        Return list of found products (in dictionaries)

        bbox: {'minlat': -0.5, 'maxlat': 0.5, 'westlon': 359.5, 'eastlon': 0.5}
        """

        assert all(k in bbox for k in ('minlat','maxlat','westlon','eastlon')), (
            "Expected 'bbox' with keys: 'minlat','maxlat','westlon','eastlon'"
        )

        contains = False if 'intersect' in match else True

        assert bbox_ref in COORDS_REF.keys()
        assert bbox['westlon'] < bbox['eastlon'], bbox
        if bbox_ref == 'C0':
            bbox = bbox.copy()
            _wl = bbox['westlon']
            _el = bbox['eastlon']
            assert _wl < _el
            bbox['westlon'] = _wl+360 if _wl < 0 else _wl
            bbox['eastlon'] = _el+360 if _el < 0 else _el


        req = request_products(bbox,
                                self.target,
                                self.mission,
                                self.instr,
                                self.ptype,
                                contains=contains)

        result = req.json()
        log.debug(f"ODE results: {result}")

        if result['ODEResults']['Status'].lower() != 'success':
            errmsg = result['ODEResults']['Error']
            log.info('Request failed:', str(errmsg))
        else:
            self._result = result
            self._ref_coords = bbox_ref

        return self

    query = query_bbox

    def query_pid(self, product_id):
        """
        Return product_id if found
        """
        req = request_product(product_id,
                                self.target,
                                self.mission,
                                self.instr,
                                self.ptype)

        result = req.json()
        log.debug(f"ODE results: {result}")

        if result['ODEResults']['Status'].lower() != 'success':
            errmsg = result['ODEResults']['Error']
            log.info('Request failed:', str(errmsg))
        else:
            self._result = result

        return self


    def _count(self):
        """
        Number of products found
        """
        if self._result is None:
            return None
        try:
            cnt = int(self._result['ODEResults']['Count'])
        except:
            print(self._result)
            raise
        return cnt

    def parse(self):
        """
        Parse ODE results into a table/dataframe with columns from METADATA
        """
        if not self._result or not self._count():
            return None
        log.debug("Search results:", self._result)

        if self._count() > 1:
            products = self._result['ODEResults']['Products']['Product']
        else:
            products = [self._result['ODEResults']['Products']['Product']]
        assert isinstance(products, list), "Was expecting 'list', got '{}' instead".format(type(products))
        if not products:
            print("No products found")
            return None

        products_output = ODEProducts()
        for i,product in enumerate(products):
            _meta = readout_product_meta(product, metadata=METADATA)
            _fprint = readout_product_footprint(product, self._ref_coords)

            files_obj = readout_product_files(product)

            _pobj = find_product_file(product_files=files_obj,
                                       product_type='product_image',
                                       descriptors=datasets.descriptors(self.dataset))
            _pfile = _pobj['URL']
            _psize = _pobj['KBytes']

            try:
                _lobj = find_product_file(product_files=files_obj,
                                           product_type='product_label',
                                           descriptors=datasets.descriptors(self.dataset))
                _lfile = _lobj['URL']
            except KeyError as err:
                _lfile = None

            try:
                _bobj = find_product_file(product_files=files_obj,
                                           product_type='browse_image',
                                           # descriptors=DESCRIPTORS[self.instr])
                                           descriptors=datasets.descriptors(self.dataset))
                _bfile = _bobj['URL']
            except KeyError as err:
                _bfile = None

            _dout = _meta
            _dout['geometry'] = _fprint
            _dout['image_url'] = _pfile
            _dout['image_kbytes'] = _psize
            _dout['label_url'] = _lfile
            _dout['browse_url'] = _bfile

            # Apply filters to product-ID (see npt.dataset._datasets)
            filters = datasets.filters(self.dataset)
            if (not filters) or select_product(_dout, filters['product-id']):
                products_output.append(_dout)

        log.info("{} products found".format(len(products_output)))
        return products_output


def select_product(meta, filter_rule):
    """
    The filter rules are (<regex>,<bool>), return (bool) if regex is/not in 'meta'
    """
    import re
    expression, present = filter_rule
    regex = re.compile(expression, re.IGNORECASE)
    product_id = meta['id']
    _match = regex.match(product_id)
    return bool(_match) is present


def request_product(product_id, target:str=None, host:str=None,
                    instr:str=None, ptype:str=None):
    """
    Return results for 'product_id' (if found)
    """
    payload = _payload_request(target, host, instr, ptype)

    payload.update({ 'pdsid': product_id })

    return _request_payload(payload)


def request_products(bbox:dict,
                     target:str=None, host:str=None,
                     instr:str=None, ptype:str=None,
                     contains:bool=False):
    """
    Assemble query-URL and (GET) request ODE server.

    ODE gets longitudes in the range [0:360]. Target, host, instr, ptype are as
    in 'npt.datasets.list'.

    bbox = {
        'minlat': <-65:65>,
        'maxlat': <-65:65>,
        'westlon': <0:360>,
        'eastlon': <0:360>
    }
    """
    payload = _payload_request(target, host, instr, ptype)

    payload.update(dict(
        minlat=bbox['minlat'],
        maxlat=bbox['maxlat'],
        westlon=bbox['westlon'],
        eastlon=bbox['eastlon']
    ))

    if contains:
        payload.update({'loc':'o'})

    return _request_payload(payload)


def _request_payload(payload):
    """
    Return result from (GET) request 'payload'
    """
    return requests.get(API_URL, params=payload)


def _payload_request(target:str, host:str, instr:str, ptype:str):
    """
    Assemble query-URL and (GET) request ODE server.
    """
    #    https://oderest.rsl.wustl.edu/live2/?
    # target=mars&
    # query=product&
    # results=fpcm&
    # output=XML&
    # pt=RDRV11&
    # iid=HiRISE&
    # ihid=MRO&
    payload = dict(
        query='product',
        results='fmpc',
        output='JSON',
        loc='f'
    )
    if target:
        payload.update({'target':target})
    if host:
        payload.update({'ihid':host})
    if instr:
        payload.update({'iid':instr})
        if ptype:
            payload.update({'pt':ptype})

    #payload.update({'pretty':True})
    return payload


# USED by 'parse_products'
def readout_product_files(product_json):
    product_files = product_json['Product_files']['Product_file']
    return product_files


# USED by 'parse_products'
def readout_product_footprint(product_json, coords_ref):
    # 'Footprint_geometry' and 'Footprint_C0_geometry' may contain 'GEOMETRYCOLLECTION'
    # when the footprint cross the meridian in "c180" or "c0" frames
    #product_geom = request.json()['ODEResults']['Products']['Product']['Footprint_geometry']
    #product_geom = request.json()['ODEResults']['Products']['Product']['Footprint_C0_geometry']
    product_geom = product_json[COORDS_REF[coords_ref]]
    return product_geom


# USED by 'parse_products'
def readout_product_meta(product_json:dict, metadata=METADATA) -> dict:
    """
    Return "{k:v}" filtered by 'metadata', plus 'id, mission, inst, type' fields

    Example:
        product_json = {}
    """
    product = {}
    # <pdsid>ESP_011712_1820_COLOR</pdsid>
    product['id'] = product_json['pdsid']
    # <ihid>MRO</ihid>
    product['mission'] = product_json['ihid']
    # <iid>HIRISE</iid>
    product['inst'] = product_json['iid']
    # <pt>RDRV11</pt>
    product['type'] = product_json['pt']

    try:
        for key in metadata:
            product[key] = product_json.get(key, None)
    except Exception as err:
        print(product_json)
        raise

    return product


# USED by 'parse_products'
def find_product_file(product_files:list, product_type:str, descriptors:dict) -> dict:
    """
    Return product from 'product_files' matching the "key,value" defined in 'descriptors[product_type]'
    (See npt.datasets._datasets)
    """
    desc_key, desc_val = descriptors[product_type]
    is_val_regex = desc_val.strip()[-1]=='*'
    desc_val_token = desc_val[:-1].strip()
    _foo = (lambda pf:
            pf[desc_key] == desc_val
            if not is_val_regex
            else desc_val_token in pf[desc_key]
            )
    pfl = list(filter(_foo, product_files))
    multiple_matches = "I was expecting one Product matching ptype '{}' but got '{}' in {}"
    assert len(pfl) == 1, multiple_matches.format(product_type, len(pfl), product_files)
    return pfl[0]



from npt.utils import filenames
from npt.utils import geojson

write_feature_json = geojson.write_feature_json

def split_label_from_image(img_filename):
    """
    Write PDS Label content from image with '.LBL' extension
    """
    label = read_image_header(img_filename)
    lbl_filename = filenames.change_extension(img_filename, 'lbl')
    with open(lbl_filename, 'w') as fp:
        fp.write(label.read())

    print("PDS LABEL file '{}' created.".format(lbl_filename))
    return lbl_filename


def read_image_header(img_filename):
    """
    Return PDS LABEL (text) from PDS IMAGE
    """
    import io
    header = io.StringIO()
    with open(img_filename, 'r') as fp:
        try:
            for line in fp:
                if '\x00' in line:
                    break
                else:
                    header.write(line)
        except:
            pass

    header.seek(0)
    return header



import versioneer
from setuptools import setup

setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    entry_points = {
        'console_scripts': ['npt=npt.cli:app'],
    }
)



import os

from . import log

def change_extension(filename, new_ext):
    """
    Change file extension
    """
    fn = '.'.join(filename.split('.')[:-1] + [new_ext])
    return fn

def change_dirname(filename, new_dir):
    """
    Change file directory path ("dirname")
    """
    bn = os.path.basename(filename)
    fn = os.path.join(new_dir, bn)
    return fn

def insert_preext(filename, sub_ext):
    """
    Insert a pre-extension before file extension

    E.g:    'file.csv' -> 'file.xyz.csv'
    """
    fs = filename.split('.')
    fs.insert(-1, sub_ext)
    fn = '.'.join(fs)
    return fn



from . import log

def string_2_dict(bbox):
    """
    Convert bbox in comma-sep string to bbox dictionary

    Input bbox (string) is as 'minlat,maxlat,westlon,eastlon'
    """
    _lbl = ['minlat','maxlat','westlon','eastlon']
    _bbx = [float(c) for c in bbox.split(',')]
    bounding_box = {k:v for k,v in zip(_lbl,_bbx)}
    return bounding_box



from npt import log
from . import sh

from npt.utils import raster

_dstools = {
    'pds2isis': {
        'ctx': sh.wrap('mroctx2isis'),
        # Test
        'test': sh.wrap(['echo','`hostname`'])
    },
}


def pds2isis(filename_in, filename_out, dataset='ctx'):
    assert dataset in _dstools['pds2isis'], "Dataset '{}' not supported.".format(dataset)
    foo = _dstools['pds2isis'][dataset]
    res = foo(FROM=filename_in, TO=filename_out)
    sh.log(res)
    return res

def init_spice(filename):
    spiceinit = sh.wrap('spiceinit')
    res = spiceinit(FROM=filename)
    sh.log(res)
    return res

def isis2tiff(filename_in, filename_out, cog=False):
    # return isissh.isis2std(FROM=filename_in, TO=filename_out, FORMAT='TIFF')
    return raster.to_tiff(filename_in, filename_out, format_in='ISIS', cog=cog)

def jpeg2tiff(filename_in, filename_out, cog=False):
    # return isissh.isis2std(FROM=filename_in, TO=filename_out, FORMAT='TIFF')
    return raster.to_tiff(filename_in, filename_out, format_in='JP2OpenJPEG', cog=cog)
